@@ -10,10 +10,10 @@ from typing import Optional, Callable, Sequence
 
 from dask.base import tokenize
 
-from flowsaber.context import context, config
+from flowsaber.core.context import context, config as flow_config
 from flowsaber.core.cache import get_cache_cls, CacheInvalidError, Cache
 from flowsaber.core.env import Env, EnvCreator
-from flowsaber.core.executor import get_executor, Executor
+from flowsaber.core.executor import Executor
 from flowsaber.core.target import File, Stdout, Stdin, END, Skip
 from flowsaber.utility.logtool import get_logger
 from flowsaber.utility.utils import change_cwd, class_deco, TaskOutput, Data, capture_local
@@ -30,6 +30,20 @@ class BaseTask(FlowComponent):
         self.num_out = num_out
         self._dependencies: Optional[dict] = None
 
+    def __call__(self, *args, **kwargs) -> TaskOutput:
+        task = self.copy_new(*args, **kwargs)
+        up_flow = context.up_flow
+        assert task not in up_flow._tasks
+        up_flow._tasks.setdefault(task, {})
+
+        return task._output
+
+    def copy_new(self, *args, **kwargs):
+        new = super().copy_new(*args, **kwargs)
+        new.initialize_output()
+        new.register_graph(new._input.queues)
+        return new
+
     def initialize_input(self, *args, **kwargs) -> Consumer:
         super().initialize_input(*args, **kwargs)
         channels = list(args) + list(kwargs.values())
@@ -37,12 +51,6 @@ class BaseTask(FlowComponent):
         self._dependencies = {}
         return self._input
 
-    def add_dependency(self, name: str, channel: Channel):
-        assert name not in self._dependencies
-        self._dependencies[name] = channel
-        self._input.add_channel(channel, consumer=self)
-        self.register_graph([self._input.queues[-1]])
-
     def initialize_output(self) -> TaskOutput:
         self._output = tuple(Channel(task=self) for i in range(self.num_out))
         if len(self._output) == 1:
@@ -64,19 +72,11 @@ class BaseTask(FlowComponent):
                 g.node(src_name, shape=shape, **kwargs)
             g.edge(src_name, self.identity_name)
 
-    def copy_new(self, *args, **kwargs):
-        new = super().copy_new(*args, **kwargs)
-        new.initialize_output()
-        new.register_graph(new._input.queues)
-        return new
-
-    def __call__(self, *args, **kwargs) -> TaskOutput:
-        task = self.copy_new(*args, **kwargs)
-        up_flow = context.up_flow
-        assert task not in up_flow._tasks
-        up_flow._tasks.setdefault(task, {})
-
-        return task._output
+    def add_dependency(self, name: str, channel: Channel):
+        assert name not in self._dependencies
+        self._dependencies[name] = channel
+        self._input.add_channel(channel, consumer=self)
+        self.register_graph([self._input.queues[-1]])
 
     async def execute(self, **kwargs):
         await super().execute(**kwargs)
@@ -161,45 +161,17 @@ class Task(BaseTask):
     def __getattr__(self, item):
         return getattr(self.config, item)
 
-    def skip(self, skip_fn: Callable):
-        assert callable(skip_fn)
-        self.skip_fn = skip_fn
-
-    def clean(self):
-        if hasattr(self, 'cache'):
-            self.cache.persist()
-
-    @property
-    def task_hash(self):
-        # get the task_hash of this task defined by the real source code of self.run
-        fn = self.run
-        # for wrapped func, use __source_func__ as a link
-        while hasattr(fn, '__source_func__'):
-            fn = fn.__source_func__
-        code = fn.__code__.co_code
-        annotations = getattr(fn, '__annotations__', None)
-
-        return tokenize(code, annotations)
-
-    @property
-    def input_hash_source(self) -> dict:
-        return {}
-
-    @property
-    def task_key(self):
-        return f"{self.__class__.__name__}-{self.task_hash}"
-
     def copy_new(self, *args, **kwargs):
         new: Task = super().copy_new(*args, **kwargs)
         # 4. update user defined config
         cls_name = new.__class__.__name__
         # update if has self's class name
-        if hasattr(config, cls_name) and isinstance(config[cls_name], dict):
-            new.config.update(config[cls_name])
+        if hasattr(flow_config, cls_name) and isinstance(flow_config[cls_name], dict):
+            new.config.update(flow_config[cls_name])
         # update if has self's base classes
         for base in self.__class__.__mro__:
-            if base in config:
-                new.config.update(config[base])
+            if base in flow_config:
+                new.config.update(flow_config[base])
 
         # set task_key and working directory
         # task key is the unique identifier of the task and task' working directory, cache, run_key_lock
@@ -249,13 +221,6 @@ class Task(BaseTask):
             task_state = scheduler.get_state(self)
             await task_state
 
-    async def enqueue_output(self, res):
-        if not isinstance(res, Skip):
-            logger.debug(f"{self} send {res} to output channel.")
-            await self._output.put(res)
-        else:
-            logger.debug(f"{self} skipped put output by SKIP")
-
     def create_run_args(self, data: Data) -> BoundArguments:
         # 1. build BoundArgument
         len_args = len(self._input_args)
@@ -268,24 +233,21 @@ class Task(BaseTask):
         run_args.apply_defaults()
         return run_args
 
-    @property
-    def cache(self) -> Cache:
-        create_cache = partial(get_cache_cls, self.config.cache_type, task=self)
-        caches = context.setdefault('__caches__', defaultdict(create_cache))
-        return caches[self.task_key]
+    async def need_skip(self, bound_args: BoundArguments) -> bool:
+        if self.skip_fn:
+            if inspect.iscoroutine(self.skip_fn):
+                return await self.skip_fn(**bound_args.arguments)
+            else:
+                return self.skip_fn(**bound_args.arguments)
+        else:
+            return False
 
-    @property
-    def executor(self) -> Executor:
-        executor_type = self.config.executor
-        executors = context.setdefault('__executors__', {})
-        if executor_type not in executors:
-            executors[executor_type] = get_executor(executor_type, config=self.config)
-
-        return executors[executor_type]
-
-    def run_key_lock(self, run_key: str):
-        locks = context.setdefault('__run_key_locks__', defaultdict(asyncio.Lock))
-        return locks[run_key]
+    async def enqueue_output(self, res):
+        if not isinstance(res, Skip):
+            logger.debug(f"{self} send {res} to output channel.")
+            await self._output.put(res)
+        else:
+            logger.debug(f"{self} skipped put output by SKIP")
 
     async def handle_input(self, run_args: BoundArguments, **kwargs):
         logger.debug(f"{self} start handle_input()")
@@ -319,7 +281,7 @@ class Task(BaseTask):
                         # clean task will not contain _xxx like attribute
                         # should maintain input/run specific information in self.run_info dict
                         clean_task = self.copy_clean()
-                        res = await self.executor.run(clean_task.run, **run_args.arguments)
+                        res = await self.executor.run(clean_task.run, **run_args.arguments, **self.config.submit_kwargs)
                     else:
                         # print(f"{self}  use cache")
                         pass
@@ -373,6 +335,15 @@ class Task(BaseTask):
                     f.hash = new_hash
         return run_args
 
+    async def update_run_info(self, data: BoundArguments):
+        pass
+
+    def run(self, *args, **kwargs):
+        """
+        This function should be stateless
+        """
+        raise NotImplementedError
+
     async def check_run_output(self, res, is_cache: bool = False):
         # make sure each File's checksum didn't change
         if is_cache:
@@ -386,28 +357,54 @@ class Task(BaseTask):
                         raise CacheInvalidError(msg)
         return res
 
-    async def update_run_info(self, data: BoundArguments):
-        pass
+    @property
+    def task_key(self):
+        return f"{self.__class__.__name__}-{self.task_hash}"
 
-    async def need_skip(self, bound_args: BoundArguments) -> bool:
-        if self.skip_fn:
-            if inspect.iscoroutine(self.skip_fn):
-                return await self.skip_fn(**bound_args.arguments)
-            else:
-                return self.skip_fn(**bound_args.arguments)
-        else:
-            return False
+    @property
+    def task_hash(self):
+        # get the task_hash of this task defined by the real source code of self.run
+        fn = self.run
+        # for wrapped func, use __source_func__ as a link
+        while hasattr(fn, '__source_func__'):
+            fn = fn.__source_func__
+        code = fn.__code__.co_code
+        annotations = getattr(fn, '__annotations__', None)
 
-    def run(self, *args, **kwargs):
-        """
-        This function should be stateless
-        """
-        raise NotImplementedError
+        return tokenize(code, annotations)
+
+    @property
+    def cache(self) -> Cache:
+        create_cache = partial(get_cache_cls, self.config.cache_type, task=self)
+        caches = context.setdefault('__caches__', defaultdict(create_cache))
+        return caches[self.task_key]
+
+    @property
+    def executor(self) -> Executor:
+        """Note: The executor can not use threads, otherwise functions relies on `os.getcwd()` may change at any time"""
+        assert '__executor__' in context.__dict__, "Executor not in flowsaber.context"
+        return context['__executor__']
+
+    @property
+    def input_hash_source(self) -> dict:
+        return {}
 
     def run_key(self, input_hash):
         assert self.task_key is not None and Path(self.workdir).is_dir()
         return str(Path(self.workdir, input_hash))
 
+    def run_key_lock(self, run_key: str):
+        locks = context.setdefault('__run_key_locks__', defaultdict(asyncio.Lock))
+        return locks[run_key]
+
+    def skip(self, skip_fn: Callable):
+        assert callable(skip_fn)
+        self.skip_fn = skip_fn
+
+    def clean(self):
+        if hasattr(self, 'cache'):
+            self.cache.persist()
+
 
 class ShellTask(Task):
     FUNC_PAIRS = [('command', 'run')]
@@ -454,9 +451,6 @@ class ShellTask(Task):
 
         return new
 
-    def command(self, *args, **kwargs) -> str:
-        raise NotImplementedError
-
     async def update_run_info(self, data: BoundArguments):
         # run user defined function and get the true bash commands
         await super().update_run_info(data)
@@ -488,6 +482,9 @@ class ShellTask(Task):
             'stdin': stdin
         })
 
+    def command(self, *args, **kwargs) -> str:
+        raise NotImplementedError
+
     def run(self, *args, **kwargs):
         """
         This function should be stateless
@@ -501,7 +498,6 @@ class ShellTask(Task):
         running_path = Path(run_key)
         pubdirs = self.config.get_pubdirs()
         with change_cwd(running_path) as path:
-
             # 2. run in environment, stdout and stderr are separated, stdout of other commands are redirected
             run_env_cmd = env.gen_cmd(str(cmd))
             stdout_file = (running_path / Path(f".__run__.stdout")).open('w')
@@ -588,12 +584,6 @@ class EnvTask(ShellTask):
             )
         return new
 
-    @property
-    def input_hash_source(self) -> dict:
-        return {
-            'env_hash': self.env_creator.hash
-        }
-
     def initialize_output(self) -> TaskOutput:
         self._output = ConstantChannel(task=self)
         return self._output
@@ -606,6 +596,12 @@ class EnvTask(ShellTask):
         self.env = env
         return env
 
+    @property
+    def input_hash_source(self) -> dict:
+        return {
+            'env_hash': self.env_creator.hash
+        }
+
 
 # decorator to make Task
 task = class_deco(Task, 'run')
